/* 
 * Linux kernel module implementing K575
 * device driver.
 *
 * written by Matthias D. Tews 1996
 * 
 */

#include <linux/version.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/ioctl.h>
#include <asm/irq.h>

#include <K575.h>
/* Auf Matthias Tews Anweisung am 1.3.96 eingefuegt (3 Zeilen #include) */
#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/version.h>

char kernel_version[] = UTS_RELEASE;

/* interrupt handling follows */

static int interrupt_mode = 0;	/* are interrupt enabled ? */
static int irq = 0;
static unsigned cnt = 0;
struct task_struct *K575_Task = NULL;

/* interrupt mode 1: 1kHz, 50% time */
  /*  changed by marki for Linux 2.0.*, was */
  /* void IrqHandler1 (int _irq, struct pt_regs *pt)  */  
 
void
IrqHandler1 (int _irq, void *dev_id, struct pt_regs *pt)
{
  if (irq != _irq)
    {
      printk ("K575 interrupt: something's wrong\n");
      return;
    }
  if (K575_Task != NULL)
    {
      switch (cnt % 2)
	{
	case 0:
	  K575_Task->counter = -16;
	  if (current == K575_Task)
	    need_resched = 1;
	  break;
	case 1:
	  K575_Task->counter = 600;	/* 6 seconds */
	  if (current != K575_Task)
	    need_resched = 1;
	  if (K575_Task->state != TASK_RUNNING)
	    K575_Task->state = TASK_RUNNING;
	  break;
	}
    }
  cnt++;
}

int
InterruptMode (unsigned mode)
{
  void (*handler) (int irq, void *dev_id,struct pt_regs * p);

  switch (mode)
    {
    case 0:
      CNT_Init ();
      return 0;
    case 1:
      handler = IrqHandler1;
      break;
    default:
      return -EINVAL;
    }

  if (irq)
    goto FOUND;

  irq = 2;
    /*  changed by marki for Linux 2.0.*, was */
  /*  if (request_irq (irq, IrqHandler1, 0, "K575") == 0) */  
  if (request_irq (irq, handler, 0, "K575", NULL) == 0)
    goto FOUND;

  irq = 9;
    /*  changed by marki for Linux 2.0.*, was */
  /*  if (request_irq (irq, IrqHandler1, 0, "K575") == 0) */  
  if (request_irq (irq, handler, 0, "K575", NULL) == 0)
    goto FOUND;

  irq = 3;
    /*  changed by marki for Linux 2.0.*, was */
  /*  if (request_irq (irq, IrqHandler1, 0, "K575") == 0) */  
  if (request_irq (irq, handler, 0, "K575", NULL) == 0)
    goto FOUND;

  irq = 5;
    /*  changed by marki for Linux 2.0.*, was */
  /*  if (request_irq (irq, IrqHandler1, 0, "K575") == 0) */  
  if (request_irq (irq, handler, 0, "K575", NULL) == 0)
    goto FOUND;

  irq = 0;
  printk ("no irq found for K575\n");
  return -EBUSY;

FOUND:
  printk ("use irq %d at %lu\n", irq, jiffies);
  CNT_SetIrqLevel (irq);
  SetByte (0xE0, 0x82);
  enable_irq (irq);

  switch (interrupt_mode = mode)
    {
    case 1:
      CNT_SetCounter (0, 1000);
      CNT_SetCounter (1, 1000);
      CNT_SetCounter (2, 1000);
      break;
    }
  return 0;
}

/* here is the ultimative ioctl function for the K575 */
static int
K575_ioctl (struct inode *inode, struct file *filp,
	    unsigned int iocmd, unsigned long ioarg)
{
  int majnr = MAJOR (inode->i_rdev);
  int retval;
  unsigned val;
  K575_ConvertStruct ConV;
  unsigned channel;

  enable_irq (irq);
  if (majnr != K575_MAJOR)
    return -EACCES;

  switch (iocmd)
    {
    case K575_AOM1_SetVoltage:
    case K575_DIO_SelectOut:
    case K575_DIO_SelectIn:
    case K575_DIO_SetByte:
    case K575_DIO_SetBit:
    case K575_DIO_UnsetBit:
    case K575_DIO_GetBit:
    case K575_DIO_GetByte:
    case K575_POKE:
    case K575_PEAK:
    case K575_CNT_GetCounter:
    case K575_CNT_SetCounter:
    case K575_CNT_InterruptMode:
      retval = verify_area (VERIFY_READ, (unsigned *) ioarg, sizeof (unsigned));

      if (retval)
	return retval;

      memcpy_fromfs (&val, (unsigned *) ioarg, sizeof (unsigned));
    }

  /* often used in the next "switch", so calculated here */
  channel = val >> 16;
  val &= 0xFFFF;

  switch (iocmd)
    {
    case K575_DIO_GetBit:
    case K575_DIO_GetByte:
    case K575_AMM1A_Value:
    case K575_AMM1A_Ready:
    case K575_PEAK:
    case K575_CNT_GetCounter:
      retval = verify_area (VERIFY_WRITE, (unsigned *) ioarg, sizeof (unsigned));

      if (retval)
	return retval;
    }

  switch (iocmd)
    {
    case K575_AOM1_SetVoltage:
      if (channel > 4 || val > 4096)
	return -EINVAL;
      AOM1_SetVoltage (2, channel, val);
      break;
    case K575_AOM1_Out:
      AOM1_Out ();
      break;
    case K575_DIO_SelectOut:
      if (val > 3)
	return -EINVAL;
      DIO_SelectOut (val);
      break;
    case K575_DIO_SelectIn:
      if (val > 3)
	return -EINVAL;
      DIO_SelectIn (val);
      break;
    case K575_DIO_SetByte:
      if (channel > 3 || val > 0xFFu)
	return -EINVAL;
      DIO_SetByte (channel, val);
      break;
    case K575_DIO_SetBit:
      if (val > 31)
	return -EINVAL;
      DIO_SetBit (val);
      break;
    case K575_DIO_UnsetBit:
      if (val > 31)
	return -EINVAL;
      DIO_UnsetBit (val);
      break;
    case K575_DIO_GetBit:
      if (val > 31)
	return -EINVAL;
      val = DIO_GetBit (val);
      memcpy_tofs ((unsigned *) ioarg, &val, sizeof (unsigned));

      break;
    case K575_DIO_GetByte:
      if (val > 3)
	return -EINVAL;
      val = DIO_GetByte (val);
      memcpy_tofs ((unsigned *) ioarg, &val, sizeof (unsigned));

      break;
    case K575_AMM1A_Convert:
      retval = verify_area (VERIFY_READ, (K575_ConvertStruct *) ioarg, sizeof (K575_ConvertStruct));
      if (retval)
	return retval;
      memcpy_fromfs (&ConV, (K575_ConvertStruct *) ioarg, sizeof (K575_ConvertStruct));
      if (ConV.channel < 0 || ConV.channel > 15)
	return -EINVAL;
      val = ConV.gain;
      if ((int) val < 0)
	(int) val = -(int) val;
      if (val != 1 && val != 2 && val != 5 && val != 10 &&
	  val != 20 && val != 50 && val != 100)
	return -EINVAL;
      if (ConV.filter != 2 && ConV.filter != 100)
	return -EINVAL;
      AMM1A_Convert (ConV.channel, ConV.gain, ConV.filter);
      break;
    case K575_AMM1A_ConvertDiff:
      retval = verify_area (VERIFY_READ, (K575_ConvertStruct *) ioarg, sizeof (K575_ConvertStruct));
      if (retval)
	return retval;
      memcpy_fromfs (&ConV, (K575_ConvertStruct *) ioarg, sizeof (K575_ConvertStruct));
      if (ConV.channel < 0 || ConV.channel > 7)
	return -EINVAL;
      val = ConV.gain;
      if ((int) val < 0)
	(int) val = -(int) val;
      if (val != 1 && val != 2 && val != 5 && val != 10 &&
	  val != 20 && val != 50 && val != 100)
	return -EINVAL;
      if (ConV.filter != 2 && ConV.filter != 100)
	return -EINVAL;
      AMM1A_ConvertDiff (ConV.channel, ConV.gain, ConV.filter);
      break;
    case K575_AMM1A_Value:
      if (AMM1A_EndOfConversion == 0)
	return -EBUSY;
      val = AMM1A_value ();
      memcpy_tofs ((unsigned *) ioarg, &val, sizeof (unsigned));

      break;
    case K575_AMM1A_Ready:
      val = AMM1A_EndOfConversion ();
      memcpy_tofs ((unsigned *) ioarg, &val, sizeof (unsigned));

      break;
    case K575_PEAK:		/* hacks for directly accessing the
				   K575 */
    case K575_POKE:
      if (val > 255 || channel < 128 || channel > 255)
	return -EINVAL;
      if (current->euid != 0)
	return -EPERM;
      if (iocmd == K575_POKE)
	SetByte (channel, val);
      else
	{
	  val = GetByte (channel);
	  memcpy_tofs ((unsigned *) (ioarg), &val, sizeof (unsigned));
	}
      return 0;
    case K575_CNT_GetCounter:
      if (val > 2)
	return -EINVAL;
      val = CNT_GetCounter (val);
      memcpy_tofs ((unsigned *) (ioarg), &val, sizeof (unsigned));

      return 0;
    case K575_CNT_SetCounter:
      if (interrupt_mode && !suser ())
	return -EPERM;
      if (channel > 2)
	return -EINVAL;
      CNT_SetCounter (channel, val);
      return 0;
    case K575_CNT_InterruptMode:
      return InterruptMode (val);
    default:
      return -ENOTTY;
    }
  return 0;
}

/* now the instalation / deinstalation procedure follow */
static int
K575_open (struct inode *inode, struct file *filp)
{
  if (filp->f_count > 1 || K575_Task != NULL)
    return -EBUSY;

  MOD_INC_USE_COUNT;
  K575_Task = current;

  return 0;
}

static void
K575_close (struct inode *inode, struct file *filp)
{
  MOD_DEC_USE_COUNT;
  K575_Task = NULL;

  if (irq)
  /*   changed by marki for Linux 2.0.*, was free_irq (irq); */
    free_irq (irq, NULL);
  irq = 0;
}

static struct file_operations K575_fops =
{
  NULL,				/* lseek */
  NULL,				/* read */
  NULL,				/* write */
  NULL,				/* readdir */
  NULL,				/* select */
  K575_ioctl,
  NULL,				/* mmap */
  K575_open,
  K575_close,
  NULL,				/* fsync */
  NULL,				/* fasync */
  NULL,				/* check_media_type */
  NULL,				/* revalidate */
};

int
init_module (void)
{
  int r;

  r = register_chrdev (K575_MAJOR, "K575", &K575_fops);

  if (r)
    {
      printk ("init-module failed\n");
      return -1;
    }
  printk ("K575 installed\n");

  AOM1_Init ();
  AMM1A_Init ();

  for (r = 0; r < 5; r++)
    AOM1_SetVoltage (2, r, 2 * 1024);	/* 0 Volts */
  AOM1_Out ();

  CNT_Init ();

  irq = 2;
    /*  changed by marki for Linux 2.0.*, was */
  /*  if (request_irq (irq, IrqHandler1, 0, "K575") == 0) */  
  if (request_irq (irq, IrqHandler1, 0, "K575", NULL) == 0)
    goto FOUND;

  irq = 9;
    /*  changed by marki for Linux 2.0.*, was */
  /*  if (request_irq (irq, IrqHandler1, 0, "K575") == 0) */  
  if (request_irq (irq, IrqHandler1, 0, "K575", NULL) == 0)
    goto FOUND;

  irq = 3;
    /*  changed by marki for Linux 2.0.*, was */
  /*  if (request_irq (irq, IrqHandler1, 0, "K575") == 0) */  
  if (request_irq (irq, IrqHandler1, 0, "K575", NULL) == 0)
    goto FOUND;

  irq = 5;
  /*  changed by marki for Linux 2.0.*, was */
  /*  if (request_irq (irq, IrqHandler1, 0, "K575") == 0) */  
  if (request_irq (irq, IrqHandler1, 0, "K575", NULL) == 0) 
    goto FOUND;

  return 0;
FOUND:
  printk ("MAIN IRQ= %d\n", irq);
  enable_irq (irq);

  return 0;
}

void
cleanup_module (void)
{
  int r;

  if (irq)
    /*   changed by marki for Linux 2.0.*, was free_irq (irq); */
    free_irq (irq, NULL);
  irq = 0;

  r = unregister_chrdev (31, "K575");
  if (r)
    printk ("remove-module failed\n");
  else
    printk ("K575 removed\n");
}
